/*
 * Hash.cpp
 *
 *  Created on: Jul 24, 2010
 *      Author: marco
 */

#include <iostream>
#include <iterator>
#include "Hash.h"

std::ostream& operator<<(std::ostream& os, const solHashed& s) {
	(os) << "hash_val: " << s.hash << " index: " << s.hash % 1200007UL
			<< " iter: " << s.iter;
	return os;
}
/*
 * Implementation of an universal hashing
 */
HashForTabu::HashForTabu(Vertex n, Color c) {
	n_vertices = n;
	n_colors = c;
	current_hash_value = 0;
	hashnum_vertices.clear();
	hashnum_vertices.resize(n);
	Longint a_i;
	for (unsigned int i = 0; i < n; i++) {
		/* LARGE_PRIME: w=1200007UL
		 * NEXT_PRIME:  m=1200061UL
		 * \varphi^{-1}=0.618033 inverse of golden ratio
		 * 741645=floor(0.618033*LARGE_PRIME
		 * QVAL: 741645 ((int) floor(0.5 * (sqrt(5.0) - 1.0) * LARGE_PRIME)) */
		/* a \in \{0,m-1}^k
		 */
		a_i = ((741645 * (i+1)) % LARGE_PRIME);
		hashnum_vertices[i] = (a_i * (i+1)) % NEXT_PRIME;
	}
	value_per_color.resize(n_colors);

	//cmpr = new compare_sum_per_colors();
	init_hash_table();

}

HashForTabu::~HashForTabu() {
	for (size_hash i = 0; i < NEXT_PRIME; i++) {
		delete hash_table[i]; // = NULL;
	}
}


void HashForTabu::init_hash_table() {
	hash_table.resize(NEXT_PRIME);
	for (size_hash i = 0; i < NEXT_PRIME; i++) {
		hash_table[i] = NULL;
	}
}

Longint HashForTabu::compute_hash_value(const Coloring &key) {
	current_hash_value = 0;
	Longint b_j;
	for (int j = 0; j < n_colors; j++) {
		VertexSet::const_iterator pt;
		value_per_color[j] = 0;
		for (pt = key.partitions[j].begin(); pt != key.partitions[j].end(); pt++)
			value_per_color[j] += hashnum_vertices[*pt];
		//assert(value_per_color[j]>=0);
		//current_hash_value += long_cast_logS(value_per_color[c]);
		b_j = (741645 * value_per_color[j]) % LARGE_PRIME;
		current_hash_value += b_j * value_per_color[j]; // % NEXT_PRIME (this is done later)
	}
	//cmpr->sort(ordered_value_per_color);
	//	std::vector<pair<Color,Longint> * >::iterator p;
	//	for (p =ordered_value_per_color.begin(); p!=ordered_value_per_color.end(); p++)
	//		cout<<"["<<(*p)->first<<" "<<(*p)->second<<"]"<<endl;
	assert(current_hash_value>0);
	return current_hash_value;
}

Longint HashForTabu::compute_hash_value_incrementally(const Coloring &st,
		ColMove &mv) {
	//mv: 0-node 1-new-col,
	Color c_old = st.col[mv.seq[0]];
	Vertex v = mv.seq[0];
	Color c_new = mv.seq[1];
	Longint delta;

	Longint A_old, A_new, B_old, B_new;
	A_old = value_per_color[c_old];
	A_new = value_per_color[c_old] - hashnum_vertices[v];
	B_old = value_per_color[c_new];
	B_new = value_per_color[c_new] + hashnum_vertices[v];

	assert(A_old>=0);
	assert(B_old>=0);
	assert(A_new>=0);
	assert(B_new>=0);

	delta = -((741645 * A_old) % LARGE_PRIME) * A_old + ((741645
			* A_new) % LARGE_PRIME) * A_new - ((741645 * B_old) % LARGE_PRIME)
			* B_old + ((741645 * B_new) % LARGE_PRIME) * B_new;

	assert(delta+current_hash_value>0);
	return delta+current_hash_value;
}


void HashForTabu::update_data(const Coloring &st,
		ColMove &mv) {
	//mv: 0-node 1-new-col,
	Color c_old = st.col[mv.seq[0]];
	Vertex v = mv.seq[0];
	Color c_new = mv.seq[1];

	value_per_color[c_new] += hashnum_vertices[v];
	value_per_color[c_old] -= hashnum_vertices[v];
}

Bool HashForTabu::insert_in_hash_table(Longint tt) {
	size_hash index = current_hash_value % NEXT_PRIME;
#ifdef TEST_HASH
	cout<<"insert_in_hash_table"<<endl;
	cout << "hash: "<<current_hash_value<< " index: " << index << " iter: " << tt << endl;
#endif
	if (hash_table[index] == NULL) {
		std::list<solHashed> *pt_l;
		pt_l = new std::list<solHashed>();
		pt_l->push_back(solHashed(current_hash_value, tt));
		hash_table[index] = pt_l;
		return true;
	} else {
		list<solHashed>::iterator p;
		for (p = (*hash_table[index]).begin(); p != (*hash_table[index]).end(); ++p) {
			//TODO:
			if ((*p).hash == current_hash_value) {
#ifdef TEST_HASH
				cout<<"Already seen"<<" " << (*p) << endl;
#endif
				(*p).iter = tt;
				return false;
			}
		}
		(*hash_table[index]).push_back(solHashed(current_hash_value, tt));
		return true;
	}
}



Bool HashForTabu::is_in_hash_table(Longint hash_value) {
	size_hash index = hash_value % NEXT_PRIME;
#ifdef TEST_HASH
	cout<<"is_in_hash_table"<<endl;
	cout << "hash: "<<hash_value<< " index: " << index<<endl;
#endif
	if (hash_table[index] == NULL) {
		return false;
	} else {
		list<solHashed>::iterator p;
		for (p = (*hash_table[index]).begin(); p != (*hash_table[index]).end(); ++p) {
			//TODO:
			if ((*p).hash == hash_value) {
				return true;
			}
		}
		return false;
	}
}
