/*
 * Parameter.h
 *
 *  Created on: Jul 22, 2010
 *      Author: marco
 */

#ifndef PARAMETER_H_
#define PARAMETER_H_

#include <cstdio>
#include <iostream>
#include <fstream>
#include <map>
#include "Definitions.h"
#include "Utilities.h"
/** set of parameters */

/** possible parameter types */
enum ParamType {
	PARAMTYPE_BOOL = 0, /**< bool values: TRUE or FALSE */
	PARAMTYPE_INT = 1, /**< integer values */
	PARAMTYPE_LONGINT = 2, /**< long integer values */
	PARAMTYPE_REAL = 3, /**< real values */
	PARAMTYPE_CHAR = 4, /**< characters */
	PARAMTYPE_STRING = 5,
	PARAMTYPE_SHORTINT = 6 /**< short integer values */
/**< strings: arrays of characters */
};
typedef enum ParamType PARAMTYPE;

typedef struct Param PARAM; /**< single parameter */
typedef struct ParamData PARAMDATA; /**< locally defined parameter specific data */

/** information method for changes in the parameter
 *
 *  Method is called if the parameter was changed through a paramsetSetXxx() call
 *  (which is called by setXxxParam()).
 *  It will not be called, if the parameter was changed directly by changing the value
 *  in the memory location.
 *
 *  input:
 *                :  main data structure
 *    param           : the changed parameter (already set to its new value)
 */

/** data for bool parameters */
struct boolParam {
	Bool* valueptr; /**< pointer to store the current parameter value, or NULL */
	Bool curvalue; /**< stores the current parameter value if it is not stored in *valueptr */
	Bool defaultvalue; /**< default value of the parameter */
};
typedef struct boolParam BOOLPARAM;

/** data for int parameters */
struct ShortintParam {
	Shortint* valueptr; /**< pointer to store the current parameter value, or NULL */
	Shortint curvalue; /**< stores the current parameter value if it is not stored in *valueptr */
	Shortint defaultvalue; /**< default value of the parameter */
	Shortint minvalue; /**< minimum value for parameter */
	Shortint maxvalue; /**< maximum value for parameter */
};
typedef struct ShortintParam SHORTINTPARAM;


/** data for int parameters */
struct IntParam {
	Int* valueptr; /**< pointer to store the current parameter value, or NULL */
	Int curvalue; /**< stores the current parameter value if it is not stored in *valueptr */
	Int defaultvalue; /**< default value of the parameter */
	Int minvalue; /**< minimum value for parameter */
	Int maxvalue; /**< maximum value for parameter */
};
typedef struct IntParam INTPARAM;

/** data for Longint parameters */
struct LongintParam {
	Longint curvalue; /**< stores the current parameter value if it is not stored in *valueptr */
	Longint defaultvalue; /**< default value of the parameter */
	Longint minvalue; /**< minimum value for parameter */
	Longint maxvalue; /**< maximum value for parameter */
	Longint* valueptr; /**< pointer to store the current parameter value, or NULL */
};
typedef struct LongintParam LONGINTPARAM;

/** data for Real parameters */
struct RealParam {
	Real curvalue; /**< stores the current parameter value if it is not stored in *valueptr */
	Real defaultvalue; /**< default value of the parameter */
	Real minvalue; /**< minimum value for parameter */
	Real maxvalue; /**< maximum value for parameter */
	Real* valueptr; /**< pointer to store the current parameter value, or NULL */
};
typedef struct RealParam REALPARAM;

/** data for char parameters */
struct CharParam {
	char* valueptr; /**< pointer to store the current parameter value, or NULL */
	char* allowedvalues; /**< array with possible parameter values, or NULL if not restricted */
	char curvalue; /**< stores the current parameter value if it is not stored in *valueptr */
	char defaultvalue; /**< default value of the parameter */
};
typedef struct CharParam CHARPARAM;

/** data for char* parameters */
struct StringParam {
	char** valueptr; /**< pointer to store the current parameter value, or NULL */
	char* curvalue; /**< stores the current parameter value if it is not stored in *valueptr */
	char* defaultvalue; /**< default value of the parameter */
};
typedef struct StringParam STRINGPARAM;

/** single parameter */
struct Param {
	union {
		BOOLPARAM boolparam; /**< data for bool parameters */
		SHORTINTPARAM shortintparam; /**< data for int parameters */
		INTPARAM intparam; /**< data for int parameters */
		LONGINTPARAM longintparam; /**< data for Longint parameters */
		REALPARAM realparam; /**< data for Real parameters */
		CHARPARAM charparam; /**< data for char parameters */
		STRINGPARAM stringparam; /**< data for char* parameters */
	} data;
	char* name; /**< name of the parameter */
	char* desc; /**< description of the parameter */
	PARAMDATA* paramdata; /**< locally defined parameter specific data */
	bool isadvanced; /**< is this parameter an advanced parameter? */
	PARAMTYPE paramtype; /**< type of this parameter */
};

struct strCmp {
	bool operator()(const char* s1, const char* s2) const {
		return strcmp(s1, s2) < 0;
	}
};

class Parameters {

public:

	Parameters() {
		paramsetCreate();
	}
	;
	~Parameters() {
		paramsetDelete();
	}
	;

	std::map<const char *, PARAM *, strCmp> paramset; /**< hash table to store the parameters */

	/** installs default values for all parameters */
	RETCODE paramsetSetToDefault();

	/** sets the parameter to its default setting */
	RETCODE paramSetToDefault(PARAM* param /**< parameter */
	);

	/** reads parameters from a file */
	void paramsetRead(const char* filename /**< file name */
	);

	/** writes all parameters in the parameter set to a file */
	RETCODE paramsetWrite(const char* filename, /**< file name, or NULL for stdout */
	bool comments, /**< should parameter descriptions be written as comments? */
	bool onlychanged /**< should only the parameters been written, that are changed from default? */
	);

	/** creates a bool parameter, sets it to its default value, and adds it to the parameter set */
	RETCODE paramsetAddBool(const char* name, /**< name of the parameter */
	const char* desc, /**< description of the parameter */
	Bool* valueptr, /**< pointer to store the current parameter value, or NULL */
	Bool isadvanced, /**< is this parameter an advanced parameter? */
	Bool defaultvalue, /**< default value of the parameter */
	PARAMDATA* paramdata /**< locally defined parameter specific data */
	);

	/** creates a int parameter, sets it to its default value, and adds it to the parameter set */

	RETCODE paramsetAddInt(const char* name, /**< name of the parameter */
	const char* desc, /**< description of the parameter */
	Int* valueptr, /**< pointer to store the current parameter value, or NULL */
	Bool isadvanced, /**< is this parameter an advanced parameter? */
	Int defaultvalue, /**< default value of the parameter */
	Int minvalue, /**< minimum value for parameter */
	Int maxvalue, /**< maximum value for parameter */
	PARAMDATA* paramdata /**< locally defined parameter specific data */
	);

	/** creates a int parameter, sets it to its default value, and adds it to the parameter set */

	RETCODE paramsetAddShortint(const char* name, /**< name of the parameter */
	const char* desc, /**< description of the parameter */
	Shortint* valueptr, /**< pointer to store the current parameter value, or NULL */
	Bool isadvanced, /**< is this parameter an advanced parameter? */
	Shortint defaultvalue, /**< default value of the parameter */
	Shortint minvalue, /**< minimum value for parameter */
	Shortint maxvalue, /**< maximum value for parameter */
	PARAMDATA* paramdata /**< locally defined parameter specific data */
	);

	/** creates a Longint parameter, sets it to its default value, and adds it to the parameter set */

	RETCODE paramsetAddLongint(const char* name, /**< name of the parameter */
	const char* desc, /**< description of the parameter */
	Longint* valueptr, /**< pointer to store the current parameter value, or NULL */
	Bool isadvanced, /**< is this parameter an advanced parameter? */
	Longint defaultvalue, /**< default value of the parameter */
	Longint minvalue, /**< minimum value for parameter */
	Longint maxvalue, /**< maximum value for parameter */
	PARAMDATA* paramdata /**< locally defined parameter specific data */
	);

	/** creates a Real parameter, sets it to its default value, and adds it to the parameter set */

	RETCODE paramsetAddReal(const char* name, /**< name of the parameter */
	const char* desc, /**< description of the parameter */
	Real* valueptr, /**< pointer to store the current parameter value, or NULL */
	Bool isadvanced, /**< is this parameter an advanced parameter? */
	Real defaultvalue, /**< default value of the parameter */
	Real minvalue, /**< minimum value for parameter */
	Real maxvalue, /**< maximum value for parameter */
	PARAMDATA* paramdata /**< locally defined parameter specific data */
	);

	/** creates a char parameter, sets it to its default value, and adds it to the parameter set */

	RETCODE paramsetAddChar(const char* name, /**< name of the parameter */
	const char* desc, /**< description of the parameter */
	char* valueptr, /**< pointer to store the current parameter value, or NULL */
	Bool isadvanced, /**< is this parameter an advanced parameter? */
	char defaultvalue, /**< default value of the parameter */
	const char* allowedvalues, /**< array with possible parameter values, or NULL if not restricted */
	PARAMDATA* paramdata /**< locally defined parameter specific data */
	);

	/** creates a string parameter, sets it to its default value, and adds it to the parameter set */

	RETCODE paramsetAddString(const char* name, /**< name of the parameter */
	const char* desc, /**< description of the parameter */
	char** valueptr, /**< pointer to store the current parameter value, or NULL */
	Bool isadvanced, /**< is this parameter an advanced parameter? */
	const char* defaultvalue, /**< default value of the parameter */
	PARAMDATA* paramdata /**< locally defined parameter specific data */
	);

private:

	RETCODE paramsetCreate();

	RETCODE paramCreate(PARAM** param, /**< pointer to the parameter */
	const char* name, /**< name of the parameter */
	const char* desc, /**< description of the parameter */
	PARAMDATA* paramdata /**< locally defined parameter specific data */
	);

	RETCODE paramCreateBool(PARAM** param, /**< pointer to the parameter */
	const char* name, /**< name of the parameter */
	const char* desc, /**< description of the parameter */
	Bool* valueptr, /**< pointer to store the current parameter value, or NULL */
	Bool isadvanced, /**< is this parameter an advanced parameter? */
	Bool defaultvalue, /**< default value of the parameter */
	PARAMDATA* paramdata /**< locally defined parameter specific data */
	);

	RETCODE paramCreateInt(PARAM** param, /**< pointer to the parameter */
	const char* name, /**< name of the parameter */
	const char* desc, /**< description of the parameter */
	int* valueptr, /**< pointer to store the current parameter value, or NULL */
	Bool isadvanced, /**< is this parameter an advanced parameter? */
	int defaultvalue, /**< default value of the parameter */
	int minvalue, /**< minimum value for parameter */
	int maxvalue, /**< maximum value for parameter */
	PARAMDATA* paramdata /**< locally defined parameter specific data */
	);


	RETCODE paramCreateShortint(PARAM** param, /**< pointer to the parameter */
	const char* name, /**< name of the parameter */
	const char* desc, /**< description of the parameter */
	Shortint* valueptr, /**< pointer to store the current parameter value, or NULL */
	Bool isadvanced, /**< is this parameter an advanced parameter? */
	Shortint defaultvalue, /**< default value of the parameter */
	Shortint minvalue, /**< minimum value for parameter */
	Shortint maxvalue, /**< maximum value for parameter */
	PARAMDATA* paramdata /**< locally defined parameter specific data */
	);


	RETCODE paramCreateChar(PARAM** param, /**< pointer to the parameter */
	const char* name, /**< name of the parameter */
	const char* desc, /**< description of the parameter */
	char* valueptr, /**< pointer to store the current parameter value, or NULL */
	Bool isadvanced, /**< is this parameter an advanced parameter? */
	char defaultvalue, /**< default value of the parameter */
	const char* allowedvalues, /**< array with possible parameter values, or NULL if not restricted */
	PARAMDATA* paramdata /**< locally defined parameter specific data */
	);

	/** creates a string parameter, and sets its value to default */
	RETCODE paramCreateString(PARAM** param, /**< pointer to the parameter */
	const char* name, /**< name of the parameter */
	const char* desc, /**< description of the parameter */
	char** valueptr, /**< pointer to store the current parameter value, or NULL */
	Bool isadvanced, /**< is this parameter an advanced parameter? */
	const char* defaultvalue, /**< default value of the parameter */
	PARAMDATA* paramdata /**< locally defined parameter specific data */
	);

	/** creates a Longint parameter, and sets its value to default */
	RETCODE paramCreateLongint(PARAM** param, /**< pointer to the parameter */
	const char* name, /**< name of the parameter */
	const char* desc, /**< description of the parameter */
	Longint* valueptr, /**< pointer to store the current parameter value, or NULL */
	Bool isadvanced, /**< is this parameter an advanced parameter? */
	Longint defaultvalue, /**< default value of the parameter */
	Longint minvalue, /**< minimum value for parameter */
	Longint maxvalue, /**< maximum value for parameter */
	PARAMDATA* paramdata /**< locally defined parameter specific data */
	);
	RETCODE paramCreateReal(PARAM** param, /**< pointer to the parameter */
	const char* name, /**< name of the parameter */
	const char* desc, /**< description of the parameter */
	Real* valueptr, /**< pointer to store the current parameter value, or NULL */
	Bool isadvanced, /**< is this parameter an advanced parameter? */
	Real defaultvalue, /**< default value of the parameter */
	Real minvalue, /**< minimum value for parameter */
	Real maxvalue, /**< maximum value for parameter */
	PARAMDATA* paramdata /**< locally defined parameter specific data */
	);

	RETCODE paramsetAdd(PARAM* param /**< parameter to add */
	);

	const char* paramtypeGetName(PARAMTYPE paramtype /**< type of parameter */
	);

	RETCODE paramsetGetBool(const char* name, /**< name of the parameter */
	Bool* value /**< pointer to store the parameter */
	);

	/** gets the value of an existing bool parameter */
	RETCODE paramsetGetbool(const char* name, /**< name of the parameter */
	bool* value /**< pointer to store the parameter */
	);

	/** gets the value of an existing int parameter */

	RETCODE paramsetGetInt(const char* name, /**< name of the parameter */
	Int* value /**< pointer to store the parameter */
	);

	RETCODE paramsetGetShortint(const char* name, /**< name of the parameter */
	Int* value /**< pointer to store the parameter */
	);


	/** gets the value of an existing Longint parameter */

	RETCODE paramsetGetLongint(const char* name, /**< name of the parameter */
	Longint* value /**< pointer to store the parameter */
	);

	/** gets the value of an existing Real parameter */

	RETCODE paramsetGetReal(const char* name, /**< name of the parameter */
	Real* value /**< pointer to store the parameter */
	);

	/** gets the value of an existing char parameter */

	RETCODE paramsetGetChar(const char* name, /**< name of the parameter */
	char* value /**< pointer to store the parameter */
	);

	/** gets the value of an existing string parameter */

	RETCODE paramsetGetString(const char* name, /**< name of the parameter */
	char** value /**< pointer to store the parameter */
	);

	/** changes the value of an existing bool parameter */
	RETCODE paramsetSetBool(const char* name, /**< name of the parameter */
	Bool value /**< new value of the parameter */
	);

	/** changes the value of an existing int parameter */
	RETCODE paramsetSetInt(const char* name, /**< name of the parameter */
	int value /**< new value of the parameter */
	);


	/** changes the value of an existing int parameter */
	RETCODE paramsetSetShortint(const char* name, /**< name of the parameter */
	int value /**< new value of the parameter */
	);


	/** changes the value of an existing Longint parameter */
	RETCODE paramsetSetLongint(const char* name, /**< name of the parameter */
	Longint value /**< new value of the parameter */
	);

	/** changes the value of an existing Real parameter */
	RETCODE paramsetSetReal(const char* name, /**< name of the parameter */
	Real value /**< new value of the parameter */
	);

	/** changes the value of an existing char parameter */
	RETCODE paramsetSetChar(const char* name, /**< name of the parameter */
	char value /**< new value of the parameter */
	);

	/** changes the value of an existing string parameter */
	RETCODE paramsetSetString(const char* name, /**< name of the parameter */
	const char* value /**< new value of the parameter */
	);

	/** changes the value of an existing bool parameter */
	RETCODE paramSetBool(PARAM* param, /**< parameter */
	Bool value /**< new value of the parameter */
	);


	/** changes the value of an existing int parameter */
	RETCODE paramSetShortint(PARAM* param, /**< parameter */
	Shortint value /**< new value of the parameter */
	);


	/** changes the value of an existing int parameter */
	RETCODE paramSetInt(PARAM* param, /**< parameter */
	Int value /**< new value of the parameter */
	);

	/** changes the value of an existing Longint parameter */
	RETCODE paramSetLongint(PARAM* param, /**< parameter */
	Longint value /**< new value of the parameter */
	);

	/** changes the value of an existing Real parameter */
	RETCODE paramSetReal(PARAM* param, /**< parameter */
	Real value /**< new value of the parameter */
	);

	/** changes the value of an existing char parameter */
	RETCODE paramSetChar(PARAM* param, /**< parameter */
	char value /**< new value of the parameter */
	);

	/** changes the value of an existing string parameter */
	RETCODE paramSetString(PARAM* param, /**< parameter */
	const char* value /**< new value of the parameter */
	);

	/** returns the array of parameters */
	PARAM** paramsetGetParams();

	/** returns the number of parameters in the parameter set */
	int paramsetGetNParams();

	RETCODE paramsetParse(char* line /**< line to parse (is modified during parse, but not freed) */
	);

	RETCODE paramParseInt(PARAM* param, /**< parameter */
	char* valuestr /**< value in string format (may be modified during parse) */
	);

	RETCODE paramParseBool(PARAM* param, /**< parameter */
	char* valuestr /**< value in string format (may be modified during parse) */
	);

	RETCODE paramParseLongint(PARAM* param, /**< parameter */
	char* valuestr /**< value in string format (may be modified during parse) */
	);

	RETCODE paramParseReal(PARAM* param, /**< parameter */
	char* valuestr /**< value in string format (may be modified during parse) */
	);

	RETCODE paramParseChar(PARAM* param, /**< parameter */
	char* valuestr /**< value in string format (may be modified during parse) */
	);

	RETCODE paramParseString(PARAM* param, /**< parameter */
	char* valuestr /**< value in string format (may be modified during parse) */
	);

	RETCODE paramWrite(PARAM* param, /**< parameter */
	std::ostream &file, /**< file stream to write parameter to, or NULL for stdout  */
	Bool comments, /**< should parameter descriptions be written as comments? */
	Bool onlychanged /**< should only the parameters been written, that are changed from default? */
	);

	RETCODE paramsetDelete();
	RETCODE paramDelete(PARAM** param /**< pointer to the parameter */
	);

	PARAM* hashtableRetrieve(const char * paramname);

};

#endif /* PARAMETER_H_ */
