/* GCP version 3.5 Copyright (C) 2003 Marco Chiarandini  
 Technische Universitaet Darmstadt
 email: machud@intellektik.informatik.tu-darmstadt.de

 This file is part of GCP.

 GCP is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.

 Foobar is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with GCP; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "../gzstream/gzstream.h"
#include <cstring>
#include "Problem.h"
#include "Solution.h"

// Input Methods

Graph::Graph() {
	num_of_vertices = 0;
	num_of_edges = 0;
	adjacency_matrix = NULL;
}

double rdouble(double x, int n) {
	if (n == 1)
		return (double) (x + 0.05);
	else if (n == 2)
		return (double) (x + 0.005);
	else if (n == 3)
		return (double) (x + 0.0005);
	else
		return 0;
}

void Graph::read_DIMACS_line(char *buf) {
	switch (buf[0]) {
	case 'p': {
		char *tmp = strtok(buf, " "); // splits buf
		//cout<<tmp<<endl;
		tmp = strtok(0, " "); // and considers the second substd::std::string
		//tmp = strtok(tmp, " "); // now process for the first value
		//cout<<tmp<<endl;
		tmp = strtok(0, " ");
		num_of_vertices = atoi(tmp);
		tmp = strtok(0, " "); // and for the second one
		declared_edges = atoi(tmp);

		adjacency_matrix = new unsigned char *[num_of_vertices];
		for (Vertex v = 0; v < num_of_vertices; v++) {
			adjacency_matrix[v] = new unsigned char[num_of_vertices];
			for (Vertex w = 0; w < num_of_vertices; w++)
				adjacency_matrix[v][w] = 0;
		}
		degree.resize(num_of_vertices, 0);
		adjacency_set.resize(num_of_vertices);
		break;
	}
	case 'e': {
		char *tmp; // temporary std::std::string
		// reads the edge information and add the edge between the
		// corresponding nodes
		tmp = strtok(buf, " "); // splits buf among the spaces
		tmp = strtok(0, " "); // consider the first number
		Vertex v = atoi(tmp) - 1;
		tmp = strtok(0, " "); // and now the second one
		Vertex w = atoi(tmp) - 1;
		if (adjacency_matrix[v][w] == 0) {
			num_of_edges++;
			if (v > w) {
				Vertex h = v;
				v = w;
				w = h;
			}
			pair<Vertex, Vertex> edge;
			edge.first = v;
			edge.second = w;
			edges.push_back(edge);
		}
		adjacency_matrix[v][w] = adjacency_matrix[w][v] = 1;

		break;
	}
	case 'c':
	case ' ':
	case '\n':
		// ignore the comments and the empty lines
		break;
	default:
		//std::cerr << "Error while reading the input file " << id << std::endl
		//  << "the line " << buf << " was not understandable" << std::endl;
		break;
	}
}

void Graph::Load(const char *id, bool print_statistics) {

	uint8_t format = 0;

	if (strstr(id, ".gz\0") != NULL)
		format = GZIP_FORMAT;
	else if (strstr(id, ".b\0") != NULL)
		format = BIN_FORMAT;
	else
		format = ASCII_FORMAT;

	num_of_edges = 0;
	declared_edges = 0;// for doublechecking file definition
	switch (format) {
	case ASCII_FORMAT: {
		cout << "...read input in ASCII format..." << endl;
		std::ifstream is(id);
		if (!is.good()) {
			cerr << "\nInstance file not found\n";
			exit( EXIT_FAILURE);
		}
		char buf[255]; // line buffer
		while (!is.eof()) {
			is.getline(buf, 255); // process each line
			//printf("%s\n",buf);fflush(stdout);
			read_DIMACS_line(buf);
		}
		is.close();
		assert(num_of_edges == declared_edges);
		break;
	}
	case GZIP_FORMAT: {
		cout << "...read input in GZIP format..." << endl;
		// actually igzstream reads also non-gz files
		// TODO: this is a kludge. Maybe a better solution? is.good seems not to work...
		std::ifstream is(id);
		if (!is.good()) {
			cerr << "\nInstance file not found\n";
			exit( EXIT_FAILURE);
		}
		is.close();
		igzstream ins(id, ios::in);

		if (!ins.good()) {
			cerr
					<< "\nProblems with instance format. Gzstream cannot open it.\n";
			exit( EXIT_FAILURE);
		}
		char buf[255]; // line buffer
		while (!ins.eof()) {
			ins.getline(buf, 255); // process each line
			//printf("%s\n",buf);fflush(stdout);
			read_DIMACS_line(buf);
		}
		/* the following works also:
		 std::string line;
		 while(std::getline(is, line)){
		 std::cout << line << std::endl;
		 }*/
		ins.close();
		assert(num_of_edges == declared_edges);
		break;
	}
	case BIN_FORMAT: {
		cout << "...read input in BIN format..." << endl;
		DIMACS_bin_format in(id);
		num_of_vertices = in.Nr_vert;
		adjacency_matrix = new unsigned char *[num_of_vertices];
		Vertex v, w;
		for (v = 0; v < num_of_vertices; v++) {
			adjacency_matrix[v] = new unsigned char[num_of_vertices];
			for (w = 0; w < num_of_vertices; w++)
				adjacency_matrix[v][w] = 0;
		}
		degree.resize(num_of_vertices, 0);
		adjacency_set.resize(num_of_vertices);

		for (w = 0; w < in.Nr_vert; w++) {
			for (v = 0; v <= w; v++) {
				if (in.get_edge(w, v)) {
					//printf("e %d %d\n",w+1,v+1 );
					if (adjacency_matrix[v][w] == 0) {
						num_of_edges++;
						pair<Vertex, Vertex> edge;
						edge.first = v;
						edge.second = w;
						edges.push_back(edge);
					}
					adjacency_matrix[v][w] = adjacency_matrix[w][v] = 1;
				}
			}
		}
		assert(num_of_edges == in.Nr_edges);
		break;
	}
	default:
		cerr << "A problem occurred in Graph::Load" << endl;
		break;
	}

	/*
	exit(0);
	bool gzfile;

	const char* result = strstr(id, ".gz\0");
	if (result == NULL) {
		gzfile = FALSE;
		//printf( "Could not find '.gz' in '%s'\n", id);
	} else {
		//printf( "Found a substring: '%s'\n", result );
		gzfile = TRUE;
	}

	num_of_edges = 0; // for doublechecking file definition (counts the read edges)
	// the following is not needed since igzstream reads also non-gz files
	if (gzfile) {
		// TODO: this is a kludge. Maybe a better solution? is.good seems not to work...
		std::ifstream is(id);
		if (!is.good()) {
			cerr << "\nInstance file not found\n";
			exit( EXIT_FAILURE);
		}
		is.close();
		igzstream ins(id, ios::in);

		if (!ins.good()) {
			cerr
					<< "\nProblems with instance format. Gzstream cannot open it.\n";
			exit( EXIT_FAILURE);
		}
		char buf[255]; // line buffer
		while (!ins.eof()) {
			ins.getline(buf, 255); // process each line
			//printf("%s\n",buf);fflush(stdout);
			read_DIMACS_line(buf);
		}
		// the following works also:
		// std::string line;
		 //while(std::getline(is, line)){
		 //std::cout << line << std::endl;
		 //}
		ins.close();
	} else {
		std::ifstream is(id);
		if (!is.good()) {
			cerr << "\nInstance file not found\n";
			exit( EXIT_FAILURE);
		}
		char buf[255]; // line buffer
		while (!is.eof()) {
			is.getline(buf, 255); // process each line
			//printf("%s\n",buf);fflush(stdout);
			read_DIMACS_line(buf);
		}
		is.close();
	}
*/

	// Organize some data
	//updating vector degree
	av_degree = 0;
	max_degree = 0;
	min_degree = VERTEX_MAX;
	for (Vertex i = 0; i < num_of_vertices; i++) {
		adjacency_set[i].clear();
		degree[i] = 0;
		for (Vertex j = 0; j < num_of_vertices; j++) {
			if (Edge(i, j)) {
				degree[i]++;
				adjacency_set[i].insert(j);
			}
		}
		if (degree[i] > max_degree)
			max_degree = degree[i];
		if (degree[i] < min_degree)
			min_degree = degree[i];

		av_degree += degree[i];
	}

	av_degree /= num_of_vertices;
	density = ((float) (num_of_edges)) / (((float) (num_of_vertices) / 2)
			* (num_of_vertices - 1));

	if (print_statistics) {
		for (Vertex i = 0; i < num_of_vertices; i++) {
			//cout<<(double) degree[i]/(num_of_vertices-1)<<" ";
		}
		//cout<<endl;
		double av_degree_norm = av_degree / (num_of_vertices - 1);
		double std_degree;
		double sum_sd = 0;
		for (Vertex i = 0; i < num_of_vertices; i++) {
			sum_sd += (degree[i] - av_degree) * (degree[i] - av_degree);
		}
		std_degree = sqrt(sum_sd / num_of_vertices);

		printf("\nx %s %d %.4f %.1f %.1f %.4f %.4f %d %d\n", basename(id),
				num_of_vertices, density, av_degree, std_degree,
				av_degree_norm, (std_degree / (num_of_vertices - 1)),
				min_degree, max_degree);
		exit( EXIT_SUCCESS);
	}

}

/* OLD in C style.
 void Graph::Load(const char *id, bool print_statistics) {
 char *tmp; // temporary std::std::string
 //tmp = strtok(strdup(id.c_str()), ":");
 char buf[255]; // line buffer
 //int nn, en = 0;
 int cen = 0; // for doublechecking file definition (counts the read edges)
 ifstream is(id);

 if (!is.is_open()) {
 cerr << "Instance file not found\n";
 exit(1);
 }

 while (!is.eof()) {
 is.getline(buf, 255); // process each line
 //printf("%s\n",buf);fflush(stdout);
 switch (buf[0]) {
 case 'p': {
 char *p = (char*) (&buf);//strstr((char*)(&buf), "p");
 //if (strstr((char*) &buf, "edge") != NULL) {
 p++;
 while (*p != '\0' && *p != '\n') {
 while (*p == ' ')
 p++;
 while (*p != ' ')
 p++;
 while (*p == ' ')
 p++;
 num_of_vertices = atoi(p);
 while (*p != ' ')
 p++;
 while (*p == ' ')
 p++;
 *p = '\0';
 }
 //}
 // reads the number of nodes and creates them
 //	  tmp = strtok(buf, "edge"); // splits buf around the 'edge' token
 //	  tmp = strtok(0, "edge"); // and considers the second substd::std::string
 //	  tmp = strtok(tmp, " "); // now process for the first value
 //	  cout<<tmp<<endl;
 //	  nn = atoi(tmp);
 //	  tmp = strtok(0, " "); // and for the second one
 //	  en = atoi(tmp);
 //	  // let's ignore everything else
 //	  // create the matrix
 //	  num_of_vertices = nn;
 //
 adjacency_matrix = new unsigned char *[num_of_vertices];
 for (Vertex v = 0; v < num_of_vertices; v++) {
 adjacency_matrix[v] = new unsigned char[num_of_vertices];
 for (Vertex w = 0; w < num_of_vertices; w++)
 adjacency_matrix[v][w] = 0;
 }
 degree.resize(num_of_vertices, 0);
 adjacency_set.resize(num_of_vertices);
 }
 break;
 case 'e': {
 // reads the edge information and add the edge between the
 // corresponding nodes
 tmp = strtok(buf, " "); // splits buf among the spaces
 tmp = strtok(0, " "); // consider the first number
 Vertex v = atoi(tmp) - 1;
 tmp = strtok(0, " "); // and now the second one
 Vertex w = atoi(tmp) - 1;
 if (adjacency_matrix[v][w] == 0) {
 cen++;
 if (v > w) {
 Vertex h = v;
 v = w;
 w = h;
 }
 pair<Vertex, Vertex> edge;
 edge.first = v;
 edge.second = w;
 edges.push_back(edge);
 }
 adjacency_matrix[v][w] = adjacency_matrix[w][v] = 1;

 }
 break;
 case 'c':
 case ' ':
 case '\n':
 // ignore the comments and the empty lines
 break;
 default:
 //std::cerr << "Error while reading the input file " << id << std::endl
 //  << "the line " << buf << " was not understandable" << std::endl;
 break;
 }
 }
 is.close();
 num_of_edges = cen;

 //exit(1);
 //updating vector degree
 av_degree = 0;

 max_degree = 0;
 min_degree = INT_MAX;
 for (Vertex i = 0; i < num_of_vertices; i++) {
 adjacency_set[i].clear();
 degree[i] = 0;
 for (Vertex j = 0; j < num_of_vertices; j++) {
 if (Edge(i, j)) {
 degree[i]++;
 adjacency_set[i].insert(j);
 }
 }
 if (degree[i] > max_degree)
 max_degree = degree[i];
 if (degree[i] < min_degree)
 min_degree = degree[i];

 av_degree += degree[i];
 }

 av_degree /= num_of_vertices;
 density = ((float) (num_of_edges)) / (((float) (num_of_vertices) / 2)
 * (num_of_vertices - 1));

 if (print_statistics) {
 for (Vertex i = 0; i < num_of_vertices; i++) {
 //cout<<(double) degree[i]/(num_of_vertices-1)<<" ";
 }
 //cout<<endl;
 double av_degree_norm = av_degree / (num_of_vertices - 1);
 double std_degree;
 double sum_sd = 0;
 for (Vertex i = 0; i < num_of_vertices; i++) {
 sum_sd += (degree[i] - av_degree) * (degree[i] - av_degree);
 }
 std_degree = sqrt(sum_sd / num_of_vertices);

 printf("%d %.4f %.1f %.1f %.4f %.4f %d %d ", num_of_vertices, density,
 av_degree, std_degree, av_degree_norm, (std_degree
 / (num_of_vertices - 1)), min_degree, max_degree);
 //exit(1);
 }

 }*/

void Graph::detectDisconnectivity(Bool force_if_disconnected) {
	vector<int> vertices(num_of_vertices, 0);
	Vertex v = 0;
	while (1) {
		VertexSet::iterator p;
		for (p = adjacency_set[v].begin(); p != adjacency_set[v].end(); p++) {
			if (vertices[*p] != 2)
				vertices[*p] = 1;
		}
		vertices[v] = 2;
		Vertex crossed = 0;
		Vertex circled = 0;
		for (Vertex i = 0; i < num_of_vertices; i++) {
			if (vertices[i] > 0)
				circled++;
			if (vertices[i] == 2)
				crossed++;
			if (vertices[i] == 1)
				v = i;

		}
		if (circled == num_of_vertices)
			return;
		if (crossed == circled) {
			cout << "graph disconnected!!" << endl;
			if (!force_if_disconnected)
				exit( EXIT_FAILURE);
			/*
			 char jn = '0';
			 do {
			 cout << "Continue? (y,n)" << endl;
			 cin >> jn;
			 } while ( jn == 'y' || jn == 'n');
			 if (jn == 'n')
			 exit(EXIT_SUCCESS);
			 */
		}
	}
}

std::ostream& operator<<(std::ostream& os, const Graph& g) {
	os << "graph ";
	for (Vertex v = 0; v < g.Vertices(); v++)
		for (Vertex w = v + 1; w < g.Vertices(); w++)
			if (g.Edge(v, w))
				os << "[" << w << "," << v << "] ";
	return os;
}

Graph::~Graph() {
	if (adjacency_matrix != NULL) {
		for (Vertex v = 0; v < num_of_vertices; v++)
			delete[] adjacency_matrix[v];
		delete[] adjacency_matrix;
	}
}

