/* GCP version 6.0 Copyright (C) 2003 Marco Chiarandini  
 Technische Universitaet Darmstadt
 email: machud@intellektik.informatik.tu-darmstadt.de

 This file is part of GCP.

 GCP is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.

 Foobar is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with GCP; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include"Solver.h"

void ConstructiveHeuristics::Solve(Coloring &initial) {
	assert(p_ctrl->init == ROS
			|| p_ctrl->init == RLF
			|| p_ctrl->init == DSATUR
			|| p_ctrl->init == RLF_STATIC
			|| p_ctrl->init == RLF_REVERSE);

	CompleteColoring current(initial);
	p_runner->initialization_procedure(current, COMPLETE_REPR);
	p_ctrl->setCurrentCost(current.cost, current.CurrColors(), 0);
	initial = *( dynamic_cast<Coloring *> (&current) );
}

void RandomRestart::Solve(Coloring &initial) {
	int restart = 0;
	CompleteColoring current(initial);
	while (restart < p_ctrl->max_restart && p_ctrl->isTimeLeft()) {
		p_runner->initialization_procedure(current,COMPLETE_REPR);
		p_ctrl->setCurrentCost(current.cost, current.CurrColors(), 0);
		if (!current.feasible)
			//p_runner->decreasing_procedure(current);
			p_runner->Run(current);
		restart++;
	}


}

void Metaheuristic_Fixed::Solve(Coloring &initial) {
	//cout<<p_ctrl->getTime()<<endl;
	CompleteColoring current(initial);
	p_runner->initialization_procedure(current,COMPLETE_REPR);

	p_ctrl->setCurrentCost(current.cost, current.CurrColors(), 0);
	//  cout<<p_runner->p_in->Vertices();

	if (p_ctrl->init == 0)
		return;
	if (!current.feasible) {
		p_runner->Run(current);
	}
	initial = *( dynamic_cast<Coloring *> (&current) );
}

void Metaheuristic_Descend::Solve(Coloring &current, int representation) {
	if (representation == PARTIAL_REPR)
		SolvePartial(current);
	else if (representation == COMPLETE_REPR)
		SolveComplete(current);
}

void Metaheuristic_Descend::SolveComplete(Coloring &initial) {
	CompleteColoring current(initial);
	//cout<<p_ctrl->getTime()<<endl;
	total_iterations = 0;
	p_runner->initialization_procedure(current, COMPLETE_REPR);

	//p_runner->TryReduceGraph(current);
	p_ctrl->setCurrentCost(current.cost, current.CurrColors(), 0);
	//cout<<p_ctrl->getTime()<<endl;
	while (!MaxIterationExpired() && p_ctrl->isTimeLeft()) {
		if (current.feasible) {

			initial = *( dynamic_cast<Coloring *> (&current) );

			p_runner->decreasing_procedure(current, COMPLETE_REPR);
			//p_runner->TryReduceGraph(current);
			p_ctrl->setCurrentCost(current.cost, current.CurrColors(), 0);
		}
		p_runner->SetMaxIteration(p_ctrl->max_iter - total_iterations);
		p_runner->Run(current);
		total_iterations = total_iterations + p_runner->NumberOfIterations();
		p_ctrl->setPartialIter(total_iterations);
		if (current.cost != 0)
			break;
	}
	p_ctrl->setPartialIter(total_iterations);
}


void Metaheuristic_Descend::SolvePartial(Coloring &initial) {
	PartialColoring current(initial);
	//cout<<p_ctrl->getTime()<<endl;
	total_iterations = 0;
	p_runner->initialization_procedure(current, PARTIAL_REPR);

	//p_runner->TryReduceGraph(current);
	p_ctrl->setCurrentCost(current.cost, current.CurrColors(), 0);

	while (!MaxIterationExpired() && p_ctrl->isTimeLeft()) {
		if (current.feasible) {
			initial = *( dynamic_cast<Coloring *> (&current) );
			//initial = current;
			p_runner->decreasing_procedure(current, PARTIAL_REPR);
			//p_runner->TryReduceGraph(current);

			p_ctrl->setCurrentCost(current.cost, current.CurrColors(), 0);
		}
		p_runner->SetMaxIteration(p_ctrl->max_iter - total_iterations);
		p_runner->Run(current);
		total_iterations = total_iterations + p_runner->NumberOfIterations();
		p_ctrl->setPartialIter(total_iterations);
		if (current.cost != 0)
			break;
	}
	p_ctrl->setPartialIter(total_iterations);
}


/**
 Sets the i-th runner managed by the solver to the passed parameter.

 @param r a pointer to a compatible runner
 @param i the runner position to set with r
 */
void ColoringMultiRunnerSolver::setRunner(ColoringRunner *r, unsigned int i) {
	assert(i < runners.size());
	runners[i] = r;
}

/**
 Adds the given runner to the list of the managed runners.

 @param r a pointer to a compatible runner to add
 */
void ColoringMultiRunnerSolver::addRunner(ColoringRunner *r) {
	runners.push_back(r);
}

/**
 Clears the vector of runners managed by the solver.
 */
void ColoringMultiRunnerSolver::ClearRunners() {
	runners.clear();
}

void ColoringMultiRunnerSolver::Solve(Coloring &initial) {
	CompleteColoring current(initial);
	//cout<<p_ctrl->getTime()<<endl;
	runners[0]->initialization_procedure(current,COMPLETE_REPR);
	p_ctrl->setCurrentCost(current.cost, current.CurrColors(), 0);
	//cout<<p_ctrl->getTime()<<endl;
	total_iterations = 0;

	fvalue past_fit;
	bool improved = true;
	while (improved && !MaxIterationExpired() && p_ctrl->isTimeLeft()) {
		improved = false;
		unsigned int i = (i + 1) % 2;
		cout << "Runner " << i << endl;
		runners[i]->update(current);
		if (current.feasible)
		{
			initial = *( dynamic_cast<Coloring *> (&current) );
			runners[i]->decreasing_procedure(current,COMPLETE_REPR);
		}
		past_fit = current.cost;
		runners[i]->SetMaxIteration(p_ctrl->max_iter - total_iterations);
		if (i == 1) {
			runners[i]->SetMaxIdleIteration(100);
			runners[i]->Run(current);
		} else {
			runners[i]->SetMaxIdleIteration(1000);
			runners[i]->Run(current);
		}
		total_iterations = total_iterations + runners[i]->NumberOfIterations();
		p_ctrl->setPartialIter(total_iterations);
		// if (past_fit > current.cost)
		cout << current.cost << endl;
		improved = true;
	}
}

/*
 void ColoringMultiRunnerSolver::Solve(Coloring &initial)
 {
 Coloring current(initial);
 //cout<<p_ctrl->getTime()<<endl;
 runners[0]->initialization_procedure(current);
 p_ctrl->setCurrentCost(current.cost,current.CurrColors(),0);
 //cout<<p_ctrl->getTime()<<endl;
 total_iterations = 0;

 fvalue past_fit;
 bool improved = true;
 while (improved)
 {
 improved = false;
 //for (unsigned int i = 0; i < runners.size(); i++)
 {
 unsigned int i =(i+1)%2;
 cout << "Runner "<<i<<endl;
 runners[0]->update(current);
 if (current.feasible)
 runners[0]->decreasing_procedure(current);
 past_fit = current.cost;
 runners[0]->SetMaxIteration(1000);
 runners[0]->Run(current);
 if (current.cost!=0)
 runners[1]->Run_procedure(current);
 total_iterations = total_iterations + runners[i]->NumberOfIterations();
 p_ctrl->setPartialIter(total_iterations);
 // if (past_fit > current.cost)
 cout<<current.cost<<endl;
 improved = true;
 }
 }
 }
 */

//------------------------------------------------------------

//HEA decreasing
void Metaheuristic_HEA::Solve(Coloring &initial) {
	SIZE = 10;
	//  cout<<p_ctrl->getTime()<<endl;
	if ((p_ctrl->init % 2) == 0)
		p_ctrl->init -= 1;
	total_iterations = 0;
	CompleteColoring current(initial);
	p_runner->initialization_procedure(current, COMPLETE_REPR);
	p_ctrl->setCurrentCost(current.cost, current.CurrColors(), 0);
	initial = *( dynamic_cast<Coloring *> (&current) );
	p_ctrl->k = current.CurrColors() - 1;
	if ((p_ctrl->init % 2) != 0)
		p_ctrl->init += 1;

	while (!MaxIterationExpired() && p_ctrl->isTimeLeft()) {
		HEA(current);
		p_ctrl->setPartialIter(total_iterations);
		if (current.feasible)
		{
			initial = *( dynamic_cast<Coloring *> (&current) );
			p_ctrl->k--;
		}
	}
	p_ctrl->setPartialIter(total_iterations);

}

//HEA dsatur but first sol given by (rlf).
void Metaheuristic_HEA_rlf::Solve(Coloring &initial) {

	SIZE = 10;
	//  cout<<p_ctrl->getTime()<<endl;
	total_iterations = 0;
	p_ctrl->init = 3;
	CompleteColoring current(initial);
	p_runner->initialization_procedure(current, COMPLETE_REPR);
	p_ctrl->setCurrentCost(current.cost, current.CurrColors(), 0);

	p_ctrl->init = 6;

	while (!MaxIterationExpired() && p_ctrl->isTimeLeft()) {
		unsigned int worst_fit_parent = 0;
		unsigned int count_cross = 0;
		unsigned int count_restart = 0;

		if (current.feasible) {
			//initial = current;
			initial = *( dynamic_cast<Coloring *> (&current) );
			p_runner->decreasing_procedure(current,COMPLETE_REPR);
			p_ctrl->k = current.CurrColors();
		}

		CompleteColoring *pop_solutions[SIZE];
		CompleteColoring *p_son = NULL;

		p_runner->SetMaxIteration(p_ctrl->ls_iter);
		for (unsigned int i = 0; i < SIZE; i++) {
			//cout<<" parents "<<i<<endl;
			pop_solutions[i] = new CompleteColoring(p_ctrl->rg, gr);
			if (i == 0) {
				*(pop_solutions[i]) = current;
			} else {
				p_runner->initialization_procedure(*pop_solutions[i], COMPLETE_REPR);
			}
			p_runner->Run(*pop_solutions[i]);
			total_iterations = total_iterations
					+ p_runner->NumberOfIterations();
			p_ctrl->setPartialIter(total_iterations);

			if (current.cost > pop_solutions[i]->cost) {
				current = *pop_solutions[i];
				//cout<<current.feasible<<"---------------------------------------"<<endl;
				if (current.feasible) {
					initial = *( dynamic_cast<Coloring *> (&current) );
					for (unsigned int j = 0; j <= i; j++) {
						delete pop_solutions[j];
					}
					//p_ctrl->setHEACrossovers(count_cross,count_restart);
					break;
					//	  p_runner->decreasing_procedure(current);
					//p_ctrl->setCurrentCost(current.cost,current.CurrColors(),0);
				}
			}
		}
		//cout<<"choose parents "<<total_iterations<<endl;


		while (!current.feasible && !MaxIterationExpired()
				&& p_ctrl->isTimeLeft()) {
			//choose parents
			int a, b;

			a = p_ctrl->rg->ran0N(SIZE - 1);
			do {
				b = p_ctrl->rg->ran0N(SIZE - 1);
			} while (b == a);

			if (pop_solutions[a]->cost >= pop_solutions[b]->cost)
				worst_fit_parent = a;
			else
				worst_fit_parent = b;

			//crossover
			//cout<<"crossover"<<endl;
			p_son = crossOver(pop_solutions[a], pop_solutions[b]);
			count_cross++;

			//local search
			//cout<<"local search "<<total_iterations<<endl;
			p_runner->Run(*p_son);
			total_iterations = total_iterations
					+ p_runner->NumberOfIterations();
			p_ctrl->setPartialIter(total_iterations);
			if (current.cost > p_son->cost) {
				current = *p_son;
				if (current.feasible) {
					initial = *( dynamic_cast<Coloring *> (&current) );
					delete p_son;
					//deallocate memory
					for (unsigned int i = 0; i < SIZE; i++) {
						delete pop_solutions[i];
					}
					break;
				}
			}

			//remove old parent
			//cout<<"remove old parent "<<total_iterations<<endl;
			delete pop_solutions[worst_fit_parent];
			pop_solutions[worst_fit_parent] = p_son;

			//cout<<"distance "<<averageDistanceSolutions(pop_solutions)<<endl;
			if (averageDistanceSolutions(pop_solutions) < 20) {

				count_restart++;
				p_ctrl->k = current.CurrColors();
				*(pop_solutions[0]) = current;
				for (unsigned int i = 1; i < SIZE; i++) {
					//pop_solutions[i] = new Coloring(p_ctrl->rg, gr);
					p_runner->initialization_procedure(*pop_solutions[i],COMPLETE_REPR);
					p_runner->SetMaxIteration(p_ctrl->ls_iter);
					p_runner->Run(*pop_solutions[i]);
					total_iterations = total_iterations
							+ p_runner->NumberOfIterations();
					p_ctrl->setPartialIter(total_iterations);
					if (current.cost > pop_solutions[i]->cost) {
						current = *pop_solutions[i];
						//deallocate memory
						if (current.feasible) {
							initial = *( dynamic_cast<Coloring *> (&current) );
							//deallocate memory
							for (unsigned int i = 0; i < SIZE; i++) {
								delete pop_solutions[i];
							}
							//p_ctrl->setHEACrossovers(count_cross,count_restart);
							break;
						}
					}
				}
			}

		}
		p_ctrl->setHEACrossovers(count_cross, count_restart);
		//p_ctrl->setPartialIter(total_iterations);
	}

}

//for Hea on a fixed color use this
void Metaheuristic_HEA_fixed::Solve(Coloring &initial) {
	SIZE = 10;

	total_iterations = 0;
	CompleteColoring current(initial);
	p_runner->initialization_procedure(current,COMPLETE_REPR);
	p_ctrl->setCurrentCost(current.cost, current.CurrColors(), 0);

	while (!MaxIterationExpired() && p_ctrl->isTimeLeft()) {
		HEA(current);
		p_ctrl->setPartialIter(total_iterations);
		if (current.feasible)
		{
			initial = *( dynamic_cast<Coloring *> (&current) );
			return;
		}
	}
}

void Metaheuristic_HEA::HEA(Coloring &current) {
	unsigned int worst_fit_parent = 0;
	unsigned int count_cross = 0;
	unsigned int count_restart = 0;

	CompleteColoring *pop_solutions[SIZE];
	CompleteColoring *p_son = NULL;
	//  int best_sol = 0;
	//    int best_fit_sol = LONG_INT;

	current.cost = FVALUE_MAX;
	current.feasible = false;

	p_runner->SetMaxIteration(p_ctrl->ls_iter);
	for (unsigned int i = 0; i < SIZE; i++) {
		//cout<<" parents "<<i<<endl;
		pop_solutions[i] = new CompleteColoring(p_ctrl->rg, gr);
		p_runner->initialization_procedure(*pop_solutions[i], COMPLETE_REPR);
		p_runner->Run(*pop_solutions[i]);
		total_iterations = total_iterations + p_runner->NumberOfIterations();
		p_ctrl->setPartialIter(total_iterations);
		if (current.cost > pop_solutions[i]->cost) {

			current = *pop_solutions[i];
			//cout<<current.feasible<<"---------------------------------------"<<endl;
			if (current.feasible) {
				for (unsigned int j = 0; j <= i; j++) {
					delete pop_solutions[j];
				}
				p_ctrl->setHEACrossovers(count_cross, count_restart);
				return;
				//	  p_runner->decreasing_procedure(current);
				//p_ctrl->setCurrentCost(current.cost,current.CurrColors(),0);
			}
		}
	}
	//cout<<"choose parents "<<total_iterations<<endl;


	while (!MaxIterationExpired() && p_ctrl->isTimeLeft()) {
		//choose parents
		int a, b;

		a = p_ctrl->rg->ran0N(SIZE - 1);
		do {
			b = p_ctrl->rg->ran0N(SIZE - 1);
		} while (b == a);

		if (pop_solutions[a]->cost >= pop_solutions[b]->cost)
			worst_fit_parent = a;
		else
			worst_fit_parent = b;

		//crossover
		//cout<<"crossover"<<endl;
		p_son = crossOver(pop_solutions[a], pop_solutions[b]);
		count_cross++;

		//local search
		//cout<<"local search "<<total_iterations<<endl;
		p_runner->Run(*p_son);
		total_iterations = total_iterations + p_runner->NumberOfIterations();
		p_ctrl->setPartialIter(total_iterations);
		if (current.cost > p_son->cost) {
			current = *p_son;
			if (current.feasible) {
				delete p_son;
				break;
			}
		}

		//remove old parent
		//cout<<"remove old parent "<<total_iterations<<endl;
		delete pop_solutions[worst_fit_parent];
		pop_solutions[worst_fit_parent] = p_son;

		//cout<<"distance "<<averageDistanceSolutions(pop_solutions)<<endl;
		if (averageDistanceSolutions(pop_solutions) < 20) {
			count_restart++;
			p_ctrl->k = current.CurrColors();
			*(pop_solutions[0]) = current;
			for (unsigned int i = 1; i < SIZE; i++) {
				//pop_solutions[i] = new Coloring(p_ctrl->rg, gr);
				p_runner->initialization_procedure(*pop_solutions[i],COMPLETE_REPR);
				p_runner->SetMaxIteration(p_ctrl->ls_iter);
				p_runner->Run(*pop_solutions[i]);
				total_iterations = total_iterations
						+ p_runner->NumberOfIterations();
				p_ctrl->setPartialIter(total_iterations);
				if (current.cost > pop_solutions[i]->cost) {
					current = *pop_solutions[i];
					//deallocate memory
					if (current.feasible) {
						for (unsigned int i = 0; i < SIZE; i++) {
							delete pop_solutions[i];
						}
						p_ctrl->setHEACrossovers(count_cross, count_restart);
						return;
					}
				}
			}
		}

	}

	//deallocate memory
	for (unsigned int i = 0; i < SIZE; i++) {
		delete pop_solutions[i];
	}
	p_ctrl->setHEACrossovers(count_cross, count_restart);

}

CompleteColoring *Metaheuristic_HEA::crossOver(CompleteColoring *par_a, CompleteColoring *par_b) {
	int ran_ass = 0;
	unsigned int parent, other_parent = 0;
	Color l;
	int max;
	int par;
	Color in_p2;

	Color k = par_a->CurrColors();

	vector<VertexSet> partitions[2];

	partitions[0] = par_a->partitions;
	partitions[1] = par_b->partitions;

	CompleteColoring *pt_son = new CompleteColoring(p_ctrl->rg, gr);
	pt_son->SetColors(k);

	for (l = 0; l < k; l++) {
		parent = (l % 2);
		max = -1;
		par = -1;

		for (Color m = 0; m < k; m++) {
			if ((partitions[parent][m].size() != 0)
					&& ((int) partitions[parent][m].size() > max)) {
				par = m;
				max = partitions[parent][m].size();
			}
		}
		if (max > 0) {
			VertexSet::iterator pt;
			pt = partitions[parent][par].begin();
			while (pt != partitions[parent][par].end()) {
				pt_son->col[*pt] = l;
				/*	      printf("\np: %ld elim: %ld ",l,pt->node);*/
				//assert((*pt)>=0 && (*pt)<gr->Vertices());

				other_parent = (parent + 1) % 2;
				if (other_parent == 0)
					in_p2 = par_a->col[*pt];
				else
					in_p2 = par_b->col[*pt];

				partitions[other_parent][in_p2].erase(*pt);
				pt++;
			}
			partitions[parent][par].clear();
		}
	}

	VertexSet::iterator pt;
	for (l = 0; l < k; l++) {
		for (pt = partitions[0][l].begin(); pt != partitions[0][l].end(); pt++) {
			ran_ass++;
			pt_son->col[*pt] = p_ctrl->rg->ran0N(k - 1);
		}
		for (pt = partitions[1][l].begin(); pt != partitions[1][l].end(); pt++) {
			ran_ass++;
			pt_son->col[*pt] = p_ctrl->rg->ran0N(k - 1);
		}
	}
	p_sm->updateColoringData(*pt_son);
	return pt_son;
}

double Metaheuristic_HEA::averageDistanceSolutions(CompleteColoring *p_pop[]) {
	Vertex sum = 0, distance, min;
	double average = 0;
	unsigned int count;

	count = 0;
	for (unsigned int out = 0; out < SIZE; out++) {
		//mean = 0;
	  min = VERTEX_MAX;

		for (unsigned int in = out + 1; in < SIZE; in++) {
			distance = p_sm->distanceLap((*p_pop[out]), (*p_pop[in]));

			sum += distance;
			if (distance < min)
				min = distance;
			count++;
		}
		//mean = ((REAL) mean)/count;
	}

	average = ((double) (2 * sum) / (SIZE * (SIZE - 1)));
	//cout<<count<<" "<<((SIZE*(SIZE-1)/2))<<" "<<average<<endl;
	return average;
}


//------------------------------------------------------------
//------------------------------------------------------------

void PenaltyFunction::Solve(Coloring &initial) {
	//cout<<p_ctrl->getTime()<<endl;

	total_iterations = 0;
	CompleteColoring current(initial);
	p_runner->initialization_procedure(current, COMPLETE_REPR);
	current.feasible = true;
	p_runner->p_ne->initAuxiliar(current.CurrColors());
	//  p_runner->TryReduceGraph(current);
	p_ctrl->setCurrentCost(current.cost, current.CurrColors(), 0);
	//cout<<p_ctrl->getTime()<<endl;
	p_runner->SetMaxIteration(p_ctrl->max_iter - total_iterations);

	current.partitions.resize(current.CurrColors() + 1);
	p_runner->Run(current);
	total_iterations = total_iterations + p_runner->NumberOfIterations();
	p_ctrl->setPartialIter(total_iterations);
	initial = *( dynamic_cast<Coloring *> (&current) );
}
//------------------------------------------------------------
//------------------------------------------------------------

void XRLFSolver::Solve(Coloring &current) {
	//cout<<p_ctrl->getTime()<<endl;
	total_iterations = 0;
	p_runner->Run(current);
	p_ctrl->setCurrentCost(current.cost, current.CurrColors(), 0);
}




//------------------------------------------------------------
//------------------------------------------------------------

void Ex_DSATURSolver::Solve(Coloring &initial) {
	CompleteColoring current(initial);
	p_runner->initialization_procedure(current, COMPLETE_REPR);
	p_ctrl->setCurrentCost(current.cost, current.CurrColors(), 0);
	initial = *( dynamic_cast<Coloring *> (&current) );
	p_runner->Run(initial);

}

