#include "TabuList.h"

TabuList::TabuList(Control *ctrl) {
	alpha = ctrl->alpha;
	tl_type = ctrl->tltype;
	tl = ctrl->tl;
	tl1 = ctrl->tl1;
	tl2 = ctrl->tl2;
	invtype = ctrl->invtype;

	switch (ctrl->tltype) {
	case 0:
		ctrl->getOutputStream() << "TabuPolicy: DEFAULT\n";
		break;
	case TL_ADAPTIVE:
		ctrl->getOutputStream() << "TabuPolicy: ADAPTIVE_" << ctrl->alpha
				<< "_" << ctrl->tl << "\n";
		break;
	case TL_INTERVAL:
		ctrl->getOutputStream() << "TabuPolicy: INTERVAL_" << ctrl->tl1 << "_"
				<< ctrl->tl2 << "\n";
		break;
	case TL_FIXED:
		ctrl->getOutputStream() << "TabuPolicy: FIXED_" << ctrl->tl << "\n";
		;
		break;
	default:
		cerr << "Error: option tltype " << ctrl->tltype << " not supported"
				<< endl;
		exit(-1);
	}
}

TabuList::TabuList(Vertex v, Color col, Random *rnd, Control *ctrl) {
	Init(v, col, rnd, ctrl);
}

void TabuList::Init(Vertex v, Color col, Random *rnd, Control *ctrl) {
	alpha = ctrl->alpha;
	tl_type = ctrl->tltype;
	tl = ctrl->tl;
	tl1 = ctrl->tl1;
	tl2 = ctrl->tl2;
	invtype = ctrl->invtype;

	p_ctrl = ctrl;
	rg = rnd;
	colors = col;
	tabu_list.clear();
	tabu_list.resize(colors * v, 0);
}

unsigned long TabuList::ComputeLength(const fvalue& conflicts) {
	switch (p_ctrl->tltype) {
	case 0:
	case TL_ADAPTIVE:
		/*
		 * minimic the
		 * 0.6 * |V^c| + RANDOM(0, 9)
		 * proposed in
		 * P. Galinier and J.K. Hao, Hybrid evolutionary algorithms for graph coloring, Journal of Combinatorial Optimization 3 (4) (1999), pp. 379397.
		 * here conflicts corresponds to the number of edges in conflicts hence as an upper bound there are twice as much vertices in conflicts.
		 * a similar dynamic scheme was proposed by:
		 * Dorne R. and Hao J. (1998a). A new genetic local search algorithm for graph coloring. In Parallel Problem Solving from Nature - PPSN V, 5th International Conference, edited by A.E. Eiben, T. Bck, M. Schoenauer, and H.P. Schwefel, vol. 1498 of Lecture Notes in Computer Science, pp. 745754. Springer Verlag, Berlin, Germany.
		 */
		return rg->ran0N(tl) + (int) (alpha * 2 * conflicts);
	case TL_INTERVAL:
		return tl1 + rg->ran0N(tl2 - tl1);
	case TL_FIXED:
		/*
		 * as in
		 * A. Hertz and D. de Werra, Using Tabu search techniques for graph coloring, Computing 39 (1987), pp. 345351.
		 */
		return tl;
	default:
		cerr << "Error: option tltype " << tl_type << " not supported" << endl;
		exit(-1);
	}
}

void TabuList::UpdateTabuList(const Coloring& st, const ColMove& mv,
		unsigned long iter) {
	unsigned long i,tabu_period;
	Vertex node;
	Color old_col, clr1, clr2;

	switch (mv.type) {
	case ONE_EX:
		tabu_period = TabuList::ComputeLength(st.cost);
		//tmp = (1-p_ctrl->alpha) * 2 * st.conflicts_set.size() * st.CurrColors();
		//tabu_period = rg->ran0N(tmp) + (int)(p_ctrl->alpha * 2 * st.conflicts_set.size()
		//* st.CurrColors());
		//tabu_period = rg->ran0N(9) + (int)(p_ctrl->alpha * st.conflicts.size());
		tabu_list[mv.seq[0] * colors + st.col[mv.seq[0]]] = iter + tabu_period;
		break;
	case I_SWAP:
		i = 2;
		while (mv.seq[i] != -1) {
			tabu_period = TabuList::ComputeLength(st.cost);
			tabu_list[mv.seq[i] * colors + mv.seq[1]] = iter + tabu_period;
			i++;
		}
		break;
	case CYCLE_EX:
		for (i = 0; i < (unsigned int) (mv.seq.size()) - 1; i++) {
			node = mv.seq[i];
			old_col = st.col[node];

			tabu_period = rg->ran0N(10) + (int) (alpha * 2 * st.cost);
			tabu_list[node * colors + old_col] = iter + tabu_period;
		}

		break;
	case PATH_EX:
		for (i = 0; i < (unsigned int) (mv.seq.size()) - 1; i++) {
			node = mv.seq[i];
			old_col = st.col[node];

			tabu_period = rg->ran0N(10) + (int) (alpha * 2 * st.cost);
			tabu_list[node * colors + old_col] = iter + tabu_period;
		}
		break;
	case KEMPE_EX:
		clr1 = mv.seq[0];
		clr2 = mv.seq[1];
		for (i = 2; i < (unsigned int) (mv.seq.size()); i++) {
			node = mv.seq[i];
			if (st.col[node] == clr2) {
				tabu_period = (unsigned int) ((float) alpha * rg->ran0N(
						st.CurrColors() * (st.CurrColors() - 1) / 2));
				tabu_list[node * colors + clr1] = iter + tabu_period;
			} else if (st.col[node] == clr1) {
				tabu_period = (unsigned int) ((float) alpha * rg->ran0N(
						st.CurrColors() * (st.CurrColors() - 1) / 2));
				tabu_list[node * colors + clr2] = iter + tabu_period;
			} else {
				cerr << "Something strange in updating tabu" << endl;
				cerr << node << " " << clr1 << " " << clr2 << " "
						<< st.col[node] << endl;
				exit(0);
			}
		}
		break;
	default:
		cerr << "something wrong in update tabu!\n";
		exit(EXIT_FAILURE);

	}
}

//------------------------------------------------------------
//------------------------------------------------------------
//------------------------------------------------------------


TabuListVertices::TabuListVertices(Vertex v, Random *rnd, Control *ctrl) {
	alpha = ctrl->alpha;
	tl_type = ctrl->tltype;
	tl = ctrl->tl;
	tl1 = ctrl->tl1;
	tl2 = ctrl->tl2;
	invtype = ctrl->invtype;

	p_ctrl = ctrl;
	rg = rnd;
	vertices = 0;
	tabu_list.clear();
	tabu_list.resize(v, 0);
}

void TabuListVertices::cleanTable() {
	for (unsigned int i = 0; i < vertices; i++)
		tabu_list[i] = 0;
}

void TabuListVertices::UpdateTabuList(const Coloring& st, const ColMove& mv,
		unsigned long iter) {
	if (mv.type == 1) {
		tabu_list[mv.seq[0]] = iter;
	} else {
		cerr << "Wrong updating " << endl;
	}
}

//------------------------------------------------------------
//------------------------------------------------------------
//------------------------------------------------------------

TabuListPart::TabuListPart(Vertex v, Color c, Random *rnd, Control *ctrl,
		Coloring *st) :
	TabuList(v, c, rnd, ctrl) {
	alpha = ctrl->alpha;
	tl_type = ctrl->tltype;
	tl = ctrl->tl;
	tl1 = ctrl->tl1;
	tl2 = ctrl->tl2;
	invtype = ctrl->invtype;

	p_st = st;
	vertices = v;
	tabu_list.clear();
	tabu_list.resize(v * v, 0);
}

bool TabuListPart::isTabu(const Vertex& v, const Color& c, unsigned long iter) {
	VertexSet::iterator pt;
	for (pt = p_st->partitions[c].begin(); pt != p_st->partitions[c].end(); pt++)
		if (tabu_list[v * vertices + (*pt)] >= iter)
			return true;
	return false;
}

void TabuListPart::UpdateTabuList(const Coloring& st, const ColMove& mv,
		unsigned long iter) {
	unsigned long tabu_period;
	Vertex node;
	Color old_col;
	VertexSet::iterator pt;
	switch (mv.type) {
	case 1:
		tabu_period = rg->ran0N(10);//+ (int)(p_ctrl->alpha * 2 * st.cost);

		for (pt = p_st->partitions[st.col[mv.seq[0]]].begin(); pt
				!= p_st->partitions[st.col[mv.seq[0]]].end(); pt++)
			tabu_list[mv.seq[0] * vertices + (*pt)] = iter + tabu_period;
		break;
	case 2:
		for (unsigned int i = 0; i < (unsigned int) (mv.seq.size()) - 1; i++) {
			node = mv.seq[i];
			old_col = st.col[node];
			tabu_period = rg->ran0N(10);
			for (pt = p_st->partitions[old_col].begin(); pt
					!= p_st->partitions[old_col].end(); pt++)
				tabu_list[node * vertices + (*pt)] = iter + tabu_period;
		}
		break;
	case 3:
		for (unsigned int i = 0; i < (unsigned int) (mv.seq.size()) - 1; i++) {
			node = mv.seq[i];
			old_col = st.col[node];
			tabu_period = rg->ran0N(10);
			for (pt = p_st->partitions[old_col].begin(); pt
					!= p_st->partitions[old_col].end(); pt++)
				tabu_list[node * vertices + (*pt)] = iter + tabu_period;
		}
		break;
	default:
		cerr << "something wrong in upadate tabu!";
		exit(1);
	}
}

void TabuListMinConf::UpdateTabuList(const Coloring& st, const ColMove& mv,
		unsigned long iter) {
	unsigned long tabu_period;
	tabu_period = tl;
	tabu_list[mv.seq[0] * colors + st.col[mv.seq[0]]] = iter + tabu_period;
}

void HistoryNovelty::UpdateTabuList(const Coloring& st, const ColMove& mv,
		unsigned long iter) {
	tabu_list[mv.seq[0] * colors + st.col[mv.seq[0]]] = iter;
}

//------------------------------------------------------------
//------------------------------------------------------------
//------------------------------------------------------------


TabuListDoubleRunner::TabuListDoubleRunner(Vertex v, Random *rnd, Control *ctrl) {
	alpha = ctrl->alpha;
	tl_type = ctrl->tltype;
	tl = ctrl->tl;
	tl1 = ctrl->tl1;
	tl2 = ctrl->tl2;
	invtype = ctrl->invtype;

	p_ctrl = ctrl;
	rg = rnd;
	vertices = v;
	tabu_list.clear();
	tabu_matrix.resize(v);
	for (Vertex i = 0; i < v; i++) {
		tabu_matrix[i].resize(v, 0);
	}
}

void TabuListDoubleRunner::cleanTable() {
	for (Vertex i = 0; i < vertices; i++)
		for (Vertex j = 0; j < vertices; j++)
			tabu_matrix[i][j] = 0;
}

void TabuListDoubleRunner::UpdateTabuList(const Coloring& st,
		const ColMove& mv, unsigned long iter) {
	//  unsigned long tabu_period;
	Vertex i, j, p, q;

	if (mv.type == 2) {
		//tabu_period = rg->ran0N(10) + (int)(p_ctrl->alpha * 2 * st.cost);
		tabu_matrix[mv.seq[0]][mv.seq[1]] = iter;
		tabu_matrix[mv.seq[1]][mv.seq[0]] = iter;
	} else if (mv.type == 6) {
		i = mv.seq[0];
		j = mv.seq[1];
		p = mv.seq[2];
		q = mv.seq[3];

		//tabu_period = rg->ran0N(10) + (int)(p_ctrl->alpha * 2 * st.cost);
		tabu_matrix[i][p] = iter;
		tabu_matrix[j][q] = iter;

	} else {
		cerr << "something wrong in upadate tabu!\n";
		exit(1);
	}
}

//------------------------------------------------------------
//------------------------------------------------------------
//------------------------------------------------------------


TabuListPerMoves::TabuListPerMoves(Vertex v, Random *rnd, Control *ctrl) {
	alpha = ctrl->alpha;
	tl_type = ctrl->tltype;
	tl = ctrl->tl;
	tl1 = ctrl->tl1;
	tl2 = ctrl->tl2;
	invtype = ctrl->invtype;

	p_ctrl = ctrl;
	rg = rnd;
	vertices = v;
	tabu_list.clear();
	tabu_matrix_list.resize(v);
	for (Vertex i = 0; i < v; i++) {
		tabu_matrix_list[i].resize(v);
		for (Vertex j = 0; j < v; j++) {
			tabu_matrix_list[i][j].clear();
		}
	}

	/*   ColMove tmp;
	 tmp.seq.push_back(1);
	 tmp.seq.push_back(2);
	 tmp.seq.push_back(1);
	 tmp.type=2;

	 ColMove tmp2;
	 tmp2.seq.push_back(3);
	 tmp2.seq.push_back(4);
	 tmp2.seq.push_back(3);
	 tmp2.type=2;

	 List(tmp).insert(tmp2);

	 cout<<(*List(tmp).find(tmp2))<<endl;
	 exit(0);*/

}

void TabuListPerMoves::cleanTable() {
	for (Vertex i = 0; i < vertices; i++)
		for (Vertex j = 0; j < vertices; j++)
			tabu_matrix_list[i][j].clear();
}

void TabuListPerMoves::UpdateTabuList(ColMove &mv, ColMove& prec_mv) {
	//  unsigned long tabu_period;
	Vertex i, j;
	//cout<<"mv "<<mv<<endl;
	//cout<<"prec_mv "<<prec_mv<<endl;
	//
	if (mv.type == 2) {
		i = mv.seq[0];
		j = mv.seq[1];
		if (i < j) {
			tabu_matrix_list[i][j].insert(mv);

		} else {
			tabu_matrix_list[j][i].insert(mv);

		}

		i = prec_mv.seq[0];
		j = prec_mv.seq[1];
		if (i < j) {
			tabu_matrix_list[i][j].insert(mv);

		} else {
			tabu_matrix_list[j][i].insert(mv);

		}
		//tabu_period = rg->ran0N(10) + (int)(p_ctrl->alpha * 2 * st.cost);
		//      tabu_matrix[mv.seq[0]][mv.seq[1]] = iter;
		//tabu_matrix[mv.seq[1]][mv.seq[0]] = iter;

	} else {
		cerr << "something wrong in upadate tabu!\n";
		exit(1);
	}
}

set<ColMove, lessMove> &TabuListPerMoves::List(const ColMove& mv) {
	//  unsigned long tabu_period;
	Vertex i, j;

	if (mv.type == 2) {
		i = mv.seq[0];
		j = mv.seq[1];
		if (i < j) {
			return tabu_matrix_list[i][j];
		} else {
			return tabu_matrix_list[j][i];
		}
		//tabu_period = rg->ran0N(10) + (int)(p_ctrl->alpha * 2 * st.cost);
		//      tabu_matrix[mv.seq[0]][mv.seq[1]] = iter;
		//tabu_matrix[mv.seq[1]][mv.seq[0]] = iter;
	} else {
		cerr
				<< "wrong in set<ColMove, lessMove> &TabuListPerMoves::List(const ColMove& mv)"
				<< endl;
		exit(0);
	}
}

void TabuListNew::UpdateTabuList(const Coloring& st, const ColMove& mv,
		unsigned long iter) {
	unsigned long tabu_period;

	tabu_period = TabuList::ComputeLength(st.cost);
	tabu_list[mv.seq[0] * colors + mv.seq[1]] = iter + tabu_period;
}

void TabuListNewOld::UpdateTabuList(const Coloring& st, const ColMove& mv,
		unsigned long iter) {
	unsigned long tabu_period;

	tabu_period = TabuList::ComputeLength(st.cost);
	tabu_list[mv.seq[0] * colors + mv.seq[1]] = iter + tabu_period;
	tabu_list[mv.seq[0] * colors + st.col[mv.seq[0]]] = iter + tabu_period;
}

void TabuListVertex::UpdateTabuList(const Coloring& st, const ColMove& mv,
		unsigned long iter) {
	unsigned long tabu_period;

	tabu_period = TabuList::ComputeLength(st.cost);
	for (Color c = 0; c < st.CurrColors(); c++)
		tabu_list[mv.seq[0] * colors + c] = iter + tabu_period;
}

//--------------------------------------------------------------------------------
//--------------------------------------------------------------------------------
//--------------------------------------------------------------------------------


TabuListVLSN::TabuListVLSN(Vertex v, Color col, Random *rnd, Control *ctrl) {
	Init(v, col, rnd, ctrl);
}

void TabuListVLSN::Init(Vertex v, Color col, Random *rnd, Control *ctrl) {
	alpha = ctrl->alpha;
	tl_type = ctrl->tltype;
	tl = ctrl->tl;
	tl1 = ctrl->tl1;
	tl2 = ctrl->tl2;
	invtype = ctrl->invtype;

	p_ctrl = ctrl;
	rg = rnd;
	colors = col;
	vertices = v;
	tabu_list.clear();
	tabu_list.resize(vertices * colors, 0);
}

void TabuListVLSN::cleanTable() {
	tabu_list.clear();
	tabu_list.resize(vertices * colors, 0);
}

bool TabuListVLSN::isTabu(const Vertex& v, const Color& c, unsigned long iter) {
	//  printf("%d %d %d %d %d",v,c,tabu_list[v*colors+c],iter);
	if (v >= vertices)
		return false;
	else
		return tabu_list[v * colors + c] >= iter;
}

void TabuListVLSN::UpdateTabuList(const Coloring& st, const ColMove& mv,
		unsigned long iter) {
	unsigned long tabu_period;
	Vertex node, clr1;
	Color new_col;

	switch (mv.type) {
	case 2:
		for (unsigned int i = 0; i < (unsigned int) (mv.seq.size()) - 1; i++) {
			node = mv.seq[i];
			if (node < vertices) {
				clr1 = mv.seq[i + 1];
				if (clr1 < vertices)
					new_col = st.col[clr1];
				else
					new_col = clr1 - vertices;
				tabu_period = ComputeLength(st.cost);
				tabu_list[node * colors + new_col] = iter + tabu_period;
			}
		}
		node = mv.seq[(mv.seq.size()) - 1];
		if (node < vertices) {
			clr1 = mv.seq[0];
			if (clr1 < vertices)
				new_col = st.col[clr1];
			else
				new_col = clr1 - vertices;
			tabu_period = ComputeLength(st.cost);
			tabu_list[node * colors + new_col] = iter + tabu_period;
		}

		break;
	default:
		cerr << "something wrong in update tabu!\n";
		exit(1);

	}
}
