/* GCP version 3.73 Copyright (C) 2003-2010 Marco Chiarandini
 University of Southern Denmark
 email: marco@imada.sdu.dk

 This file is part of GCP.

 GCP is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.

 GCP is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with GCP; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <cmath>
#include <cfloat>
#include <string>
#include<fstream>

#include <csignal>

#include <sys/time.h>
#include <sys/resource.h>

#include "Timer.h"
#include "Random.h"
#include "Utilities.h"
#include "Problem.h"
#include "bin2asc.h"
#include "Solution.h"
#include "Explorer.h"
#include "Runners.h"
#include "Solver.h"
#include "Settings.h"

#ifdef EXTERNAL_SOLVERS
#include "../Porumbel/src/porumbel.h"
#include "../MMT/malaguti.h"
#endif
//#define METAOUTPUT
//#define TEST_RUNNER
//#define TEST_LAP_DIST

using namespace std;

void got_timeout(int s) {
	cerr << "Interrupted. Time Limit exceeded!\n";
	exit( EXIT_FAILURE);
}

void check_stack_size() {
	const rlim_t kStackSize = 10L * 1024L * 1024L; // min stack size = 64 Mb
	struct rlimit rl;
	int result;

	result = getrlimit(RLIMIT_STACK, &rl);
	cout << "max stack size in bits: " << rl.rlim_cur;
	if (result == 0) {
		if (rl.rlim_cur < kStackSize) {
			rl.rlim_cur = kStackSize;
			result = setrlimit(RLIMIT_STACK, &rl);
			if (result != 0) {
				cerr << "setrlimit returned result = " << result;
			}
			cout << " ...max stack size reset to 10Mb bits: " << rl.rlim_cur;
		}
	} else {
		cerr
				<< "Could not determine stack size. Terminating. Is this a linux system? "
				<< endl;
		exit( EXIT_FAILURE);
	}
	cout << endl;
}

long int *array;

int main(int argc, char **argv) {
	check_stack_size();
	Random random(0);
	Control control(&random, argc, argv);

	Timer halt;
	halt.setAlarm((int) (control.getNumberOfTries() * control.getTimeLimit()
			+ 5), got_timeout);

	control.resetTime();

	Graph graph;
	(control.getOutputStream()) << "Reading: " << control.getInputStream();

	graph.Load(control.getInputStream(), control.print_statistics);
	(control.getOutputStream()) << " Done. Time: "
			<< (control.getTime() < 0 ? 0.0 : control.getTime()) << "\n";

#ifdef METAOUTPUT
	control.informationsReport();
#endif

	graph.detectDisconnectivity(control.force_if_disconnected);

	//Coloring *coloring = new Coloring(&random, &graph);
	while (control.triesLeft()) {

		Coloring *coloring = new Coloring(&random, &graph);
		ColoringManager *coloringManager = NULL;
		ColNeighborhoodExplorer *coloringExplorer = NULL;
		ColoringRunner *coloringRunner = NULL;
		TabuList *tabuList = NULL;
		ColoringSolver *coloringSolver = NULL;// = new ColoringSolver(&control, coloringRunner);

		int representation = COMPLETE_REPR;

		switch (control.invtype) {
		case 0:
			break;
		case INV_OLD:
			tabuList = new TabuList(&control);
			control.getOutputStream() << "InverseMove: OLD\n";
			break;
		case INV_NEW:
			tabuList = new TabuListNew(&control);
			control.getOutputStream() << "InverseMove: NEW\n";
			break;
		case INV_NEWOLD:
			tabuList = new TabuListNewOld(&control);
			control.getOutputStream() << "InverseMove: NEW-OLD\n";
			break;
		case INV_VERTEX:
			tabuList = new TabuListVertex(&control);
			control.getOutputStream() << "InverseMove: VERTEX\n";
			break;
		case VLSN:
			tabuList = new TabuListVLSN(&control);
			control.getOutputStream() << "VLSN InverseMove: OLD\n";
			break;
		default:
			cerr << "Error: option invtype " << control.invtype
					<< " not supported" << endl;
			exit(-1);
		}

		switch (control.runner) {
		case BASIC:
			coloringManager = new ColoringManager(&random, &graph, &control);
			coloringExplorer = new ColOneExExplorer(coloringManager, &random,
					&graph);
			coloringRunner = new ColoringRunner(coloringManager,
					coloringExplorer, &random, &control, &graph);
			break;
		case TABUCOL:
			coloringManager = new ColoringManager(&random, &graph, &control);
			coloringExplorer = new ColOneExExplorer(coloringManager, &random,
					&graph);
			coloringRunner = new ColoringTabuSearch(coloringManager,
					coloringExplorer, &random, &control, tabuList, &graph);
			break;
		case PARTIALCOL:
			coloringManager = new ColoringManager(&random, &graph, &control);
			coloringExplorer = new ColISwapExplorer(coloringManager, &random,
					&graph);
			coloringRunner = new ColoringPartialCol(coloringManager,
					coloringExplorer, &random, &control, tabuList, &graph);
			representation = PARTIAL_REPR;
			break;
		case TABU_HASH:
			coloringManager = new ColoringManager(&random, &graph, &control);
			coloringExplorer = new ColOneExExplorer(coloringManager, &random,
					&graph);
			coloringRunner = new ColoringTabuHashSearch(coloringManager,
					coloringExplorer, &random, &control, &graph);
			break;
		case TABU_REACTIVE:
			coloringManager = new ColoringManager(&random, &graph, &control);
			coloringExplorer = new ColOneExExplorer(coloringManager, &random,
					&graph);
			coloringRunner = new ColoringReactiveTabuSearch(coloringManager,
					coloringExplorer, &random, &control, tabuList, &graph);
			break;
			//----------------------------------------------------------------------
		case TABU_VLSN: // See version gcp-v3.5 for all the variants of VLSN tried. Explorer for VLSN in file cycle_hash.
			coloringManager = new ColoringManager(&random, &graph, &control);
			coloringExplorer
					= new ColRevRedMinOneExOrNegHashCycleOrPathExplorerII(
							coloringManager, &random, 50, &graph);
			coloringRunner = new ColoringTabuSearchWithoutAsp(coloringManager,
					coloringExplorer, &random, &control, &graph);
			break;
			//----------------------------------------------------------------------
		case SA_KEMPE:
			coloringManager = new ColManPenaltyFunction(&random, &graph,
					&control);
			coloringExplorer
					= new KempeChains(coloringManager, &random, &graph);
			coloringRunner = new SimulatedAnnealing(coloringManager,
					coloringExplorer, &random, &control, &graph);
			break;
		case GLS:
			coloringManager
					= new ColManWithPenalties(&random, &graph, &control);
			coloringExplorer = new ColOneExExplorer(coloringManager, &random,
					&graph);
			coloringRunner = new GuidedLocalSearch(coloringManager,
					coloringExplorer, &random, &control, &graph);
			break;
		case ILS:
			coloringManager = new ColoringManager(&random, &graph, &control);
			coloringExplorer = new ColOneExExplorer(coloringManager, &random,
					&graph);
			coloringRunner = new IteratedLocalSearch(coloringManager,
					coloringExplorer, &random, &control, &graph);
			break;
		case NOVELTY:
			coloringManager = new ColoringManager(&random, &graph, &control);
			coloringExplorer = new ColOneExNoveltyExplorer(coloringManager,
					&random, &graph);
			coloringRunner = new NoveltyRunner(coloringManager,
					coloringExplorer, &random, &control, &graph);
			break;
		case MIN_CONFLICT:
			coloringManager = new ColoringManager(&random, &graph, &control);
			coloringExplorer = new ColOneExMinConfExplorer(coloringManager,
					&random, &graph);
			coloringRunner = new MinConflicts(coloringManager,
					coloringExplorer, &random, &control, tabuList, &graph);
			break;
		case X_RLF_RUNNER:
			coloringManager = new ColoringManager(&random, &graph, &control);
			coloringRunner = new XRLF(coloringManager, &random, &control,
					&graph);
			break;
		case EX_DSATUR_RUNNER:
			coloringManager = new ColoringManager(&random, &graph, &control);
			coloringRunner = new Ex_DSATUR(coloringManager, &random, &control,
					&graph);
			break;

		default:
			cerr << "No runner + explorer selected\n";
			break;
		}

		if (coloringRunner == NULL)
			coloringRunner = new ColoringRunner(coloringManager, &random,
					&control, &graph);

		vector<ColoringRunner*> runners;

		if (control.max_iter == 0) { // just for historical reasons, this will never be the case
			control.max_iter = 10000 * graph.Vertices();
			control.getOutputStream() << "max_iter reset to "
					<< control.max_iter << endl;
		} else if (control.max_iter != DEFAULT_TOT_MAX_ITER) {
			control.max_iter = control.max_iter * graph.Vertices();
			control.getOutputStream() << "max_iter reset to "
					<< control.max_iter << endl;
		}

		coloringRunner->SetMaxIdleIteration(control.max_idle);
		coloringRunner->SetMaxIteration(control.max_iter);

		switch (control.solver) {
		case CONSTRUCTION_HEUR:
			coloringSolver = new ConstructiveHeuristics(&control,
					coloringRunner);
			break;
		case RANDOM_RESTART:
			coloringSolver = new RandomRestart(&control, coloringRunner);
			break;
		case FIXED_K_COL:
			coloringSolver = new Metaheuristic_Fixed(&control, coloringRunner);
			break;
		case SEQUENCE_K_COL:
			coloringSolver
					= new Metaheuristic_Descend(&control, coloringRunner);
			break;
		case HYBRID_EA: //HEA
			coloringSolver = new Metaheuristic_HEA(&control, coloringRunner,
					&graph, coloringManager);
			break;
		case PENALTY_FUNCTION: //PenaltyFunction
			coloringSolver = new PenaltyFunction(&control, coloringRunner);
			break;
		case HEA_RLF: //HEA with first solution determined by RLF and the others by dsatur
			coloringSolver = new Metaheuristic_HEA_rlf(&control,
					coloringRunner, &graph, coloringManager);
			break;
		case X_RLF_SOLVER:
			coloringSolver = new XRLFSolver(&control, coloringRunner);
			break;
		case EX_DSATUR_SOLVER:
			coloringSolver = new Ex_DSATURSolver(&control, coloringRunner);
			break;
#if EXTERNAL_SOLVERS
			case PORUMBELS_EA:
			coloringSolver = new ConstructiveHeuristics(&control,
					coloringRunner);
			break;
			case MALAGUTIS_EA:
			coloringSolver = new ConstructiveHeuristics(&control,
					coloringRunner);
			break;
#endif
		default:
			cerr << "Wrong solver setting\n";
			exit(1);
		}
		//------------------------------------------------------------
		control.beginTry();

		coloringSolver->Solve(*coloring, representation);
#ifdef EXTERNAL_SOLVERS
		if (control.solver == PORUMBELS_EA)
		{
			Porumbel(control.fullPathToInstance(),
					coloring->CurrColors(),
					(int) (control.getTimeLimit()-control.getTime()),
					control.getSeed());
		}
		else if (control.solver == MALAGUTIS_EA)
		{
			Malaguti(control.fullPathToInstance(),
					coloring->CurrColors(),
					(int) (control.getTimeLimit()-control.getTime()),
					control.getSeed());
		}
#endif
		control.endTry();
		//------------------------------------------------------------

		if (control.certificate) {
			ofstream ooo(control.os_certificate.c_str());
			//ooo<<control.real_size_solved<<endl;
			ooo << (*coloring);//ooo<<original_coloring;
			ooo.close();
			control.getOutputStream() << "Solution certificate printed in "
					<< control.os_certificate.c_str() << endl;
		}

		delete coloringSolver;
		delete coloringRunner;
		delete coloringExplorer;
		delete coloringManager;
		delete coloring;
	}
	return EXIT_SUCCESS;

}

